﻿//-------------------------------------------------------------------------------//
//Copyright (c) 2016, Vincent Bengtsson
//All rights reserved.

//Redistribution and use in source and binary forms, with or without
//modification, are permitted provided that the following conditions are met:

//1. Redistributions of source code must retain the above copyright notice, this
//   list of conditions and the following disclaimer. 
//2. Redistributions in binary form must reproduce the above copyright notice,
//   this list of conditions and the following disclaimer in the documentation
//   and/or other materials provided with the distribution.

//THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
//ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
//WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
//DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
//ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
//(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
//LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
//ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
//SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

//The views and conclusions contained in the software and documentation are those
//of the authors and should not be interpreted as representing official policies, 
//either expressed or implied, of the FreeBSD Project.
//-------------------------------------------------------------------------------//

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using System.Windows.Forms;

public class ExtendedTcpClient
{
    bool TerminatePacketListener = false;
    bool PacketListenerActive = false;

    #region Properties
    private TcpClient _tcpclient;

    /// <summary>
    /// Gets the underlying TcpClient.
    /// </summary>
    /// <remarks></remarks>
    public TcpClient Client
    {
        get { return _tcpclient; }
    }

    /// <summary>
    /// Gets or sets the owner of this class (required).
    /// </summary>
    /// <remarks></remarks>
    public Form OwnerForm { get; set; }

    /// <summary>
    /// Gets the IP address to the connected endpoint.
    /// </summary>
    /// <remarks></remarks>
    public string Address
    {
        get
        {
            if (this.Client != null && this.Client.Connected == true)
                return ((IPEndPoint)this.Client.Client.RemoteEndPoint).Address.ToString();
            return null;
        }
    }

    /// <summary>
    /// Gets the port to the connected endpoint.
    /// </summary>
    /// <remarks></remarks>
    public int Port
    {
        get
        {
            if (this.Client != null && this.Client.Connected == true)
                return ((IPEndPoint)this.Client.Client.RemoteEndPoint).Port;
            return -1;
        }
    }

    /// <summary>
    /// Gets a Boolean value indicating wether the client is connected to an endpoint or not.
    /// </summary>
    /// <remarks></remarks>
    public bool Connected
    {
        get { return this.Client != null && this.Client.Connected; }
    }

    /// <summary>
    /// Gets or sets the underlying NetworkStream of the TcpClient (automatically set when connecting).
    /// </summary>
    /// <remarks></remarks>
    public NetworkStream ClientNetworkStream { get; set; }

    #endregion

    #region Events

    #region Delegates

    public delegate void PacketReceivedEventHandler(object sender, PacketReceivedEventArgs e);

    #endregion

    #region Event Declarations

    /// <summary>
    /// Occurs when a packet was received.
    /// </summary>
    public event PacketReceivedEventHandler PacketReceived;

    #endregion

    #region Thread-safe raising
    private void ThreadSafeRaise_PacketReceived(object sender, PacketReceivedEventArgs e)
    {
        if (OwnerForm.InvokeRequired == true)
        {
            OwnerForm.Invoke((Action)(delegate {ThreadSafeRaise_PacketReceived(sender, e);}));
        }
        else
        {
            if (PacketReceived != null) { PacketReceived(sender, e); }
        }
    }
    #endregion

    #region Event Args
    public class PacketReceivedEventArgs : EventArgs
    {


        private TcpMessagePacket _packet;
        /// <summary>
        /// Gets the received packet.
        /// </summary>
        /// <remarks></remarks>
        public TcpMessagePacket Packet
        {
            get { return _packet; }
        }

        /// <summary>
        /// Initializes a new instance of the PacketReceivedEventArgs class.
        /// </summary>
        /// <param name="Packet">The packet that was received.</param>
        /// <remarks></remarks>
        public PacketReceivedEventArgs(TcpMessagePacket Packet)
        {
            this._packet = Packet;
        }
    }
    #endregion

    #endregion

    #region Constructors
    /// <summary>
    /// Initializes a new instance of the ExtendedTcpClient class.
    /// </summary>
    /// <param name="OwnerForm">The owner of this class.</param>
    public ExtendedTcpClient(Form OwnerForm)
    {
        this.OwnerForm = OwnerForm;
        this._tcpclient = new TcpClient();
    }

    /// <summary>
    /// Initializes a new instance of the ExtendedTcpClient class and connects it to the specified endpoint.
    /// </summary>
    /// <param name="OwnerForm">The owner of this class.</param>
    /// <param name="Address">The IP address of the endpoint to connect to.</param>
    /// <param name="Port">The port of the endpoint to connect to.</param>
    public ExtendedTcpClient(Form OwnerForm, string Address, int Port)
    {
        this.OwnerForm = OwnerForm;
        this._tcpclient = new TcpClient();
        this.Connect(Address, Port);
    }
    #endregion

    #region Methods
    /// <summary>
    /// Attempts to connect to the specified endpoint.
    /// </summary>
    /// <param name="Address">The IP address of the endpoint to connect to.</param>
    /// <param name="Port">The port of the endpoint to connect to.</param>
    /// <remarks></remarks>
    public void Connect(string Address, int Port)
    {
        try
        {
            this.Client.Connect(Address, Port);
            this.ClientNetworkStream = this.Client.GetStream();
            this.StartPacketListener();
        }
        catch (Exception ex)
        {
            MessageBox.Show("Couldn't connect to endpoint:" + Environment.NewLine + ex.Message + Environment.NewLine + ex.StackTrace, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
        }
    }

    /// <summary>
    /// Disconnects the client from the endpoint and disposes the underlying TcpClient.
    /// </summary>
    /// <remarks></remarks>
    public void Disconnect()
    {
        TerminatePacketListener = true; //Stop the packet listener thread.
        PacketListenerActive = false;

        //Handle the disconnection in a separate thread.
        Thread DThread = new Thread(DisconnectThread);
        DThread.IsBackground = true;
        DThread.Start();

        this.ClientNetworkStream = null;
    }

    /// <summary>
    /// (USED BY THREADS ONLY) The method that handles the disconnection of the client.
    /// </summary>
    /// <remarks></remarks>
    private void DisconnectThread()
    {
        System.Threading.Thread.Sleep(25); //Wait just a little moment so the packet listener thread can terminate.
        try { Client.GetStream().Close(5000); } catch {}
        try { Client.Close(); } catch {}
        try { Client.Client.Dispose(); } catch {}
    }

    /// <summary>
    /// Starts the thread that listens for packets (automatically started by Connect()).
    /// </summary>
    /// <remarks></remarks>
    private void StartPacketListener()
    {
        if (PacketListenerActive == true) { throw new Exception("Packet listener is already active!"); }
        Thread PacketListenerThread = new Thread(PacketListener);
        PacketListenerThread.IsBackground = true;
        PacketListenerThread.Start();
        PacketListenerActive = true;
    }

    /// <summary>
    /// Changes/sets the underlying TcpClient.
    /// </summary>
    /// <param name="TcpClient">The TcpClient to use as the new underlying TcpClient.</param>
    /// <remarks></remarks>
    public void SetNewClient(ref TcpClient TcpClient)
    {
        if (PacketListenerActive == true) { throw new Exception("Disconnect the client before setting a new one."); }
        this._tcpclient = TcpClient;
        this.ClientNetworkStream = this.Client.GetStream();
        this.StartPacketListener();
    }
    #endregion

    #region PacketListener()

    private void PacketListener()
    {
        int PacketLength = -1;
        int BytesRead = 0;
        byte[] ReceivedBytes = new byte[0];

        while(this.Client != null && this.Connected == true && TerminatePacketListener == false)
        {
            while(this.ClientNetworkStream != null && this.ClientNetworkStream.DataAvailable == true)
            {
                if(PacketLength < 0) //Read the packet length.
                {
                    if(BytesRead == 0) //Only execute if no bytes have been read yet.
                    {
                        ReceivedBytes = new byte[4]; //Converting an Integer to bytes results in 4 bytes, so we need an array of length 4.
                    }

                    BytesRead += ClientNetworkStream.Read(ReceivedBytes, BytesRead, (ReceivedBytes.Length - BytesRead));

                    if(BytesRead == 4)
                    {
                        PacketLength = BitConverter.ToInt32(ReceivedBytes, 0);
                        BytesRead = 0; //Reset the total amount of bytes read so that we can start reading the packet from the beginning.
                        Array.Clear(ReceivedBytes, 0, ReceivedBytes.Length);
                    }


                }
                else //Read the packet contents.
                {
                    if(BytesRead == 0) { ReceivedBytes = new byte[PacketLength]; }

                    BytesRead += ClientNetworkStream.Read(ReceivedBytes, BytesRead, (ReceivedBytes.Length - BytesRead));

                    if(BytesRead >= PacketLength) //Check so that we've received the whole packet.
                    {
                        TcpMessagePacket.PacketHeader PacketHeader = (TcpMessagePacket.PacketHeader)Extensions.PickByte(ref ReceivedBytes, 0); //Take out the header from the packet.
                        ThreadSafeRaise_PacketReceived(this, new PacketReceivedEventArgs(new TcpMessagePacket(ReceivedBytes, PacketHeader))); //Raise the PacketReceived event, with the newly received packet.

                        //Reset the variables so we can start receiving new packets.
                        BytesRead = 0;
                        PacketLength = -1;
                        Array.Clear(ReceivedBytes, 0, ReceivedBytes.Length);
                    }
                }
            }

            Thread.Sleep(1);
        }
    }

    #endregion
}
