﻿//-------------------------------------------------------------------------------//
//Copyright (c) 2016, Vincent Bengtsson
//All rights reserved.

//Redistribution and use in source and binary forms, with or without
//modification, are permitted provided that the following conditions are met:

//1. Redistributions of source code must retain the above copyright notice, this
//   list of conditions and the following disclaimer. 
//2. Redistributions in binary form must reproduce the above copyright notice,
//   this list of conditions and the following disclaimer in the documentation
//   and/or other materials provided with the distribution.

//THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
//ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
//WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
//DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
//ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
//(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
//LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
//ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
//SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

//The views and conclusions contained in the software and documentation are those
//of the authors and should not be interpreted as representing official policies, 
//either expressed or implied, of the FreeBSD Project.
//-------------------------------------------------------------------------------//

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Net.Sockets;

public class TcpMessagePacket
{
    #region Properties

    /// <summary>
    /// Gets or sets the header of the packet (what kind of packet this is).
    /// </summary>
    /// <remarks></remarks>
    public PacketHeader Header { get; set; }

    /// <summary>
    /// Gets or sets the contents of the packet.
    /// </summary>
    /// <remarks></remarks>
    public byte[] Contents { get; set; }

    /// <summary>
    /// Gets how many bytes the packet consists of (not including the length-prefix).
    /// </summary>
    /// <remarks></remarks>
    public int PacketLength
    {
        get { return this.Contents.Length; }
    }

    /// <summary>
    /// Gets how many bytes the packet consists of (not including the length-prefix, but including the packet header).
    /// </summary>
    /// <remarks></remarks>
    public int PacketLengthWithHeader
    {
        //Header is one byte.
        get { return this.Contents.Length + 1; }
    }

    #endregion

    #region Methods

    /// <summary>
    /// Sends a length-prefixed packet with it's header and it's contents via the specified TcpClient.
    /// </summary>
    /// <param name="TcpClient">The TcpClient to use when sending the packet.</param>
    /// <remarks></remarks>
    public void Send(ref TcpClient TcpClient)
    {
        //- Complete packet structure -
        //Byte 0-3: Packet length.
        //Byte 4  : Packet header.
        //Byte 5-?: Packet contents.
        byte[] FinalPacket = BitConverter.GetBytes(this.PacketLengthWithHeader).Concatenate(new byte[1] { (byte)this.Header }).Concatenate(this.Contents);
        TcpClient.GetStream().Write(FinalPacket, 0, FinalPacket.Length);
    }

    #endregion

    #region Constructors

    /// <summary>
    /// Initializes a new instance of the TcpMessagePacket class.
    /// </summary>
    /// <remarks></remarks>
    public TcpMessagePacket()
	{
		this.Contents = new byte[0];
	}

    /// <summary>
    /// Initializes a new instance of the TcpMessagePacket class.
    /// </summary>
    /// <param name="InitialData">The initial data to be in the packet.</param>
    /// <remarks></remarks>
    public TcpMessagePacket(byte[] InitialData)
    {
        this.Contents = InitialData;
    }

    /// <summary>
    /// Initializes a new instance of the TcpMessagePacket class.
    /// </summary>
    /// <param name="InitialData">The initial data to be in the packet.</param>
    /// <param name="Header">The header of the packet (what kind of packet this is).</param>
    /// <remarks></remarks>
    public TcpMessagePacket(byte[] InitialData, PacketHeader Header)
    {
        this.Contents = InitialData;
        this.Header = Header;
    }

    #endregion

    #region Constants

    /// <summary>
    /// A series of packet headers, used when sending packets via TCP.
    /// </summary>
    /// <remarks></remarks>
    public enum PacketHeader : byte
    {
        /// <summary>
        /// Indicates that the current packet has no specific header.
        /// </summary>
        /// <remarks></remarks>
        None = 0,
        /// <summary>
        /// Indicates that the current packet is a string (plain text).
        /// </summary>
        /// <remarks></remarks>
        PlainText
        //Keep adding rows of text here to add more headers. No need to give them any values, the compiler does that for you.
    }

    #endregion
}

#region Extensions
public static class Extensions
{
    /// <summary>
    /// Concatenate two arrays together.
    /// </summary>
    /// <typeparam name="T">The type of array to be concatenated together.</typeparam>
    /// <param name="InitialArray">The array to be in the beginning of the new array.</param>
    /// <param name="SecondaryArray">The array to be in the end of the new array.</param>
    /// <remarks></remarks>
    public static T[] Concatenate<T>(this T[] InitialArray, T[] SecondaryArray)
    {
        int ArrayStart = InitialArray.Length;
        Array.Resize(ref InitialArray, ArrayStart + SecondaryArray.Length);
        Array.Copy(SecondaryArray, 0, InitialArray, ArrayStart, SecondaryArray.Length);
        return InitialArray;
    }

    /// <summary>
    /// Returns the byte at the specified index and removes it from the array.
    /// </summary>
    /// <param name="ByteArray">The array to pick a byte from.</param>
    /// <param name="Index">The zero-based index of the byte to get.</param>
    /// <remarks></remarks>
    public static byte PickByte(ref byte[] ByteArray, int Index)
    {
        byte ReturnByte = ByteArray[Index];

        List<byte> ByteList = ByteArray.ToList();
        ByteList.RemoveAt(Index);
        ByteArray = ByteList.ToArray();

        return ReturnByte;
    }
}
#endregion
